@@ -1,6 +1,4 @@
 #!/usr/bin/env python
-# -*- coding: utf-8; mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
-# vim: fileencoding=utf-8 tabstop=4 expandtab shiftwidth=4
 #
 # Copyright 2013 Lowell Alleman
 #
@@ -52,34 +50,32 @@ for testing, patches are welcome.
 
 See the README file for an example usage of this module.
 
-This module supports Python 2.6 and later.
-
+This module supports Python 3.6 and later.
+(Support for older version was dropped in 0.9.23.)
 """
 
 import errno
-import io
 import os
 import sys
 import time
 import traceback
 import warnings
 from contextlib import contextmanager
-from logging.handlers import BaseRotatingHandler
+from logging.handlers import BaseRotatingHandler, TimedRotatingFileHandler
 
-# noinspection PyPackageRequirements
 from portalocker import LOCK_EX, lock, unlock
-from concurrent_log_handler.__version__ import __author__, __version__
+
+from concurrent_log_handler.__version__ import __author__, __version__  # noqa: F401
 
 try:
-    import pwd
     import grp
+    import pwd
 except ImportError:
     pwd = grp = None
 
 # Random numbers for rotation temp file names, using secrets module if available (Python 3.6).
 # Otherwise use `random.SystemRandom` if available, then fall back on `random.Random`.
 try:
-    # noinspection PyPackageRequirements,PyCompatibility
     from secrets import randbits
 except ImportError:
     import random
@@ -100,11 +96,11 @@ except ImportError:
 
 __all__ = [
     "ConcurrentRotatingFileHandler",
+    "ConcurrentTimedRotatingFileHandler",
 ]
 
-PY2 = False
-if sys.version_info[0] == 2:
-    PY2 = True
+has_chown = hasattr(os, "chown")
+has_chmod = hasattr(os, "chmod")
 
 
 class ConcurrentRotatingFileHandler(BaseRotatingHandler):
@@ -114,7 +110,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
     exceed the given size.
     """
 
-    def __init__(
+    def __init__(  # noqa: PLR0913
         self,
         filename,
         mode="a",
@@ -132,8 +128,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         unicode_error_policy="ignore",
         lock_file_directory=None,
     ):
-        """
-        Open the specified file and use it as the stream for logging.
+        """Open the specified file and use it as the stream for logging.
 
         :param filename: name of the log file to output to.
         :param mode: write mode: defaults to 'a' for text append
@@ -161,7 +156,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         doesn't support. Default is to ignore, i.e., drop the unusable characters.
         :param lock_file_directory: name of directory for all lock files as alternative
         living space; this is useful for when the main log files reside in a cloud synced
-        drive like Dropbox, OneDrive, Google Docs, etc, which may prevent the lock files
+        drive like Dropbox, OneDrive, Google Docs, etc., which may prevent the lock files
         from working correctly. The lock file must be accessible to all processes writing
         to a shared log, including across all different hosts (machines).
 
@@ -198,14 +193,13 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         self.umask = umask
         self._set_uid = None
         self._set_gid = None
-        self.use_gzip = True if gzip and use_gzip else False
         self._rotateFailed = False
         self.maxBytes = maxBytes
         self.backupCount = backupCount
         self.newline = newline
 
         self._debug = debug
-        self.use_gzip = True if gzip and use_gzip else False
+        self.use_gzip = bool(gzip and use_gzip)
         self.gzip_buffer = 8096
         self.maxLockAttempts = 20
 
@@ -215,6 +209,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                 "Invalid unicode_error_policy for concurrent_log_handler: "
                 "must be ignore, replace, or strict. Defaulting to ignore.",
                 UserWarning,
+                stacklevel=3,
             )
         self.unicode_error_policy = unicode_error_policy
 
@@ -223,6 +218,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                 "concurrent_log_handler parameter `delay` is now ignored and implied as True, "
                 "please remove from your config.",
                 DeprecationWarning,
+                stacklevel=3,
             )
 
         # Construct the handler with the given arguments in "delayed" mode
@@ -234,15 +230,16 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
 
         self.terminator = terminator or "\n"
 
-        if owner and os.chown and pwd and grp:
-            # noinspection PyUnresolvedReferences
+        if owner and has_chown and pwd and grp:
             self._set_uid = pwd.getpwnam(self.owner[0]).pw_uid
-            # noinspection PyUnresolvedReferences
             self._set_gid = grp.getgrnam(self.owner[1]).gr_gid
 
         self.lockFilename = self.getLockFilename(lock_file_directory)
         self.is_locked = False
 
+        # This is primarily for the benefit of the unit tests.
+        self.num_rollovers = 0
+
     def getLockFilename(self, lock_file_directory):
         """
         Decide the lock filename. If the logfile is file.log, then we use `.__file.lock` and
@@ -251,19 +248,19 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         :param lock_file_directory: name of the directory for alternative living space of lock files
         :return: the path to the lock file.
         """
-        if self.baseFilename.endswith(".log"):
-            lock_file = self.baseFilename[:-4]
-        else:
-            lock_file = self.baseFilename
-        lock_file += ".lock"
-        lock_path, lock_name = os.path.split(lock_file)
-        # hide the file on Unix and generally from file completion
-        lock_name = ".__" + lock_name
+        lock_path, lock_name = self.baseLockFilename(self.baseFilename)
         if lock_file_directory:
             self.__create_lock_directory__(lock_file_directory)
             return os.path.join(lock_file_directory, lock_name)
-        else:
-            return os.path.join(lock_path, lock_name)
+        return os.path.join(lock_path, lock_name)
+
+    @staticmethod
+    def baseLockFilename(baseFilename):
+        lock_file = baseFilename[:-4] if baseFilename.endswith(".log") else baseFilename
+        lock_file += ".lock"
+        lock_path, lock_name = os.path.split(lock_file)
+        # hide the file on Unix and generally from file completion
+        return lock_path, ".__" + lock_name
 
     @staticmethod
     def __create_lock_directory__(lock_file_directory):
@@ -280,17 +277,33 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             self._console_log("Lockfile already open in this process")
             return
         lock_file = self.lockFilename
-        self._console_log(
-            "concurrent-log-handler %s opening %s" % (hash(self), lock_file),
-            stack=False,
-        )
+        # self._console_log(
+        #     f"concurrent-log-handler {hash(self)} opening {lock_file}",
+        #     stack=False,
+        # )
 
         with self._alter_umask():
-            self.stream_lock = open(lock_file, "wb", buffering=0)
+            self.stream_lock = self.atomic_open(lock_file)
 
         self._do_chown_and_chmod(lock_file)
 
-    def _open(self, mode=None):
+    def atomic_open(self, file_path):
+        try:
+            # Attempt to open the file in "r+" mode
+            file = open(file_path, "r+", encoding=self.encoding, newline=self.newline)
+        except FileNotFoundError:
+            # If the file doesn't exist, create it atomically and open in "r+" mode
+            try:
+                fd = os.open(file_path, os.O_CREAT | os.O_EXCL | os.O_RDWR)
+                file = open(fd, "r+", encoding=self.encoding, newline=self.newline)
+            except FileExistsError:
+                # If the file was created between the first check and our attempt to create it, open it in "r+" mode
+                file = open(
+                    file_path, "r+", encoding=self.encoding, newline=self.newline
+                )
+        return file
+
+    def _open(self, mode=None):  # noqa: ARG002
         # Normally we don't hold the stream open. Only do_open does that
         # which is called from do_write().
         return None
@@ -306,8 +319,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             mode = self.mode
 
         with self._alter_umask():
-            # noinspection PyArgumentList
-            stream = io.open(
+            stream = open(
                 self.baseFilename,
                 mode=mode,
                 encoding=self.encoding,
@@ -355,26 +367,15 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         if stack:
             stack_str = ":\n" + "".join(traceback.format_stack())
         asctime = time.asctime()
-        print(
-            "[%s %s %s] %s%s"
-            % (
-                tid,
-                pid,
-                asctime,
-                msg,
-                stack_str,
-            )
-        )
+        print(f"[{tid} {pid} {asctime}] {msg}{stack_str}")
 
     def emit(self, record):
-        """
-        Emit a record.
+        """Emit a record.
 
         Override from parent class to handle file locking for the duration of rollover and write.
         This also does the formatting *before* locks are obtained, in case the format itself does
         logging calls from within. Rollover also occurs while the lock is held.
         """
-        # noinspection PyBroadException
         try:
             msg = self.format(record)
             try:
@@ -384,7 +385,9 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                     if self.shouldRollover(record):
                         self.doRollover()
                 except Exception as e:
-                    self._console_log("Unable to do rollover: %s" % (e,), stack=True)
+                    self._console_log(
+                        f"Unable to do rollover: {e}\n{traceback.format_exc()}"
+                    )
                     # Continue on anyway
 
                 self.do_write(msg)
@@ -406,69 +409,41 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         self.stream = self.do_open()
         stream = self.stream
 
-        if PY2:
-            self.do_write_py2(msg)
-        else:
-            msg = msg + self.terminator
+        msg = msg + self.terminator
+        try:
+            stream.write(msg)
+        except UnicodeError:
+            # Try to emit in a form acceptable to the output encoding
+            # The unicode_error_policy determines whether this is lossy.
             try:
+                encoding = getattr(stream, "encoding", self.encoding or "us-ascii")
+                msg_bin = msg.encode(encoding, self.unicode_error_policy)
+                msg = msg_bin.decode(encoding, self.unicode_error_policy)
                 stream.write(msg)
             except UnicodeError:
-                # Try to emit in a form acceptable to the output encoding
-                # The unicode_error_policy determines whether this is lossy.
-                try:
-                    encoding = getattr(stream, "encoding", self.encoding or "us-ascii")
-                    msg_bin = msg.encode(encoding, self.unicode_error_policy)
-                    msg = msg_bin.decode(encoding, self.unicode_error_policy)
-                    stream.write(msg)
-                except UnicodeError:
-                    # self._console_log(str(e))
-                    raise
+                # self._console_log(str(e))
+                raise
 
         stream.flush()
         self._close()
         return
 
-    # noinspection PyCompatibility,PyUnresolvedReferences
-    def do_write_py2(self, msg):
-        stream = self.stream
-        term = self.terminator
-        policy = self.unicode_error_policy
-
-        encoding = getattr(stream, "encoding", None)
-
-        # as far as I can tell, this should always be set from io.open, but just in case...
-        if not encoding:
-            if not self.encoding:
-                self._console_log(
-                    "Warning, unable to determine encoding of logging stream; assuming utf-8"
-                )
-            encoding = self.encoding or "utf-8"
-
-        if not isinstance(msg, unicode):  # noqa: F821
-            msg = unicode(msg, encoding, policy)  # noqa: F821
-
-        # Add in the terminator.
-        if not isinstance(term, unicode):  # noqa: F821
-            term = unicode(term, encoding, policy)  # noqa: F821
-        msg = msg + term
-        stream.write(msg)
-
     def _do_lock(self):
         if self.is_locked:
             return  # already locked... recursive?
         self._open_lockfile()
         if self.stream_lock:
-            for i in range(self.maxLockAttempts):
-                # noinspection PyBroadException
+            for _i in range(self.maxLockAttempts):
                 try:
                     lock(self.stream_lock, LOCK_EX)
                     self.is_locked = True
+                    # self._console_log("Acquired lock")
                     break
-                except Exception:
+                except Exception:  # noqa: S112
                     continue
             else:
                 raise RuntimeError(
-                    "Cannot acquire lock after %s attempts" % (self.maxLockAttempts,)
+                    f"Cannot acquire lock after {self.maxLockAttempts} attempts"
                 )
         else:
             self._console_log("No self.stream_lock to lock", stack=True)
@@ -478,6 +453,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             if self.is_locked:
                 try:
                     unlock(self.stream_lock)
+                    # self._console_log("Released lock")
                 finally:
                     self.is_locked = False
                     self.stream_lock.close()
@@ -486,15 +462,14 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             self._console_log("No self.stream_lock to unlock", stack=True)
 
     def close(self):
-        """
-        Close log stream and stream_lock."""
+        """Close log stream and stream_lock."""
         self._console_log("In close()", stack=True)
         try:
             self._close()
         finally:
             super(ConcurrentRotatingFileHandler, self).close()
 
-    def doRollover(self):
+    def doRollover(self):  # noqa: C901
         """
         Do a rollover, as described in __init__().
         """
@@ -507,32 +482,27 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             return
 
         # Determine if we can rename the log file or not. Windows refuses to
-        # rename an open file, Unix is inode base so it doesn't care.
+        # rename an open file, Unix is inode based, so it doesn't care.
 
         # Attempt to rename logfile to tempname:
         # There is a slight race-condition here, but it seems unavoidable
         tmpname = None
         while not tmpname or os.path.exists(tmpname):
-            tmpname = "%s.rotate.%08d" % (self.baseFilename, randbits(64))
+            tmpname = f"{self.baseFilename}.rotate.{randbits(64):08}"
         try:
             # Do a rename test to determine if we can successfully rename the log file
             os.rename(self.baseFilename, tmpname)
 
             if self.use_gzip:
                 self.do_gzip(tmpname)
-        except (IOError, OSError):
-            exc_value = sys.exc_info()[1]
-            self._console_log(
-                "rename failed.  File in use? exception=%s" % (exc_value,), stack=True
-            )
+        except OSError as e:
+            self._console_log(f"rename failed.  File in use? e={e}", stack=True)
             return
 
-        gzip_ext = ""
-        if self.use_gzip:
-            gzip_ext = ".gz"
+        gzip_ext = ".gz" if self.use_gzip else ""
 
         def do_rename(source_fn, dest_fn):
-            self._console_log("Rename %s -> %s" % (source_fn, dest_fn + gzip_ext))
+            self._console_log(f"Rename {source_fn} -> {dest_fn + gzip_ext}")
             if os.path.exists(dest_fn):
                 os.remove(dest_fn)
             if os.path.exists(dest_fn + gzip_ext):
@@ -555,8 +525,8 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
 
         do_renames = []
         for i in range(1, self.backupCount):
-            sfn = self.rotation_filename("%s.%d" % (self.baseFilename, i))
-            dfn = self.rotation_filename("%s.%d" % (self.baseFilename, i + 1))
+            sfn = self.rotation_filename(f"{self.baseFilename}.{i}")
+            dfn = self.rotation_filename(f"{self.baseFilename}.{i + 1}")
             if os.path.exists(sfn + gzip_ext):
                 do_renames.append((sfn, dfn))
             else:
@@ -574,9 +544,9 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             logFilename = self.baseFilename + ".1.gz"
             self._do_chown_and_chmod(logFilename)
 
-        self._console_log("Rotation completed")
+        self.num_rollovers += 1
+        self._console_log("Rotation completed (on size)")
 
-    # noinspection PyUnusedLocal
     def shouldRollover(self, record):
         """
         Determine if rollover should occur.
@@ -584,6 +554,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         For those that are keeping track. This differs from the standard
         library's RotatingLogHandler class. Because there is no promise to keep
         the file size under maxBytes we ignore the length of the current record.
+        TODO: should we reconsider this and make it more exact?
         """
         del record  # avoid pychecker warnings
         return self._shouldRollover()
@@ -605,28 +576,295 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             return
         out_filename = input_filename + ".gz"
 
-        with open(input_filename, "rb") as input_fh:
-            with gzip.open(out_filename, "wb") as gzip_fh:
-                while True:
-                    data = input_fh.read(self.gzip_buffer)
-                    if not data:
-                        break
-                    gzip_fh.write(data)
+        with open(input_filename, "rb") as input_fh, gzip.open(
+            out_filename, "wb"
+        ) as gzip_fh:
+            while True:
+                data = input_fh.read(self.gzip_buffer)
+                if not data:
+                    break
+                gzip_fh.write(data)
 
         os.remove(input_filename)
-        self._console_log("#gzipped: %s" % (out_filename,), stack=False)
+        self._console_log(f"#gzipped: {out_filename}", stack=False)
         return
 
     def _do_chown_and_chmod(self, filename):
-        if self._set_uid is not None and self._set_gid is not None:
+        if has_chown and self._set_uid is not None and self._set_gid is not None:
             os.chown(filename, self._set_uid, self._set_gid)
 
-        if self.chmod is not None and os.chmod:
+        if has_chmod and self.chmod is not None:
             os.chmod(filename, self.chmod)
 
 
-# Publish this class to the "logging.handlers" module so that it can be use
+class ConcurrentTimedRotatingFileHandler(TimedRotatingFileHandler):
+    """A time-based rotating log handler that supports concurrent access across
+    multiple processes or hosts (using logs on a shared network drive).
+
+    You can also include size-based rotation by setting maxBytes > 0.
+    WARNING: if you only want time-based rollover and NOT also size-based, set maxBytes=0,
+    which is already the default.
+    Please note that when size-based rotation is done, it still uses the naming scheme
+    of the time-based rotation. If multiple rotations had to be done within the timeframe of
+    the time-based rollover name, then a number like ".1" will be appended to the end of the name.
+
+    Note that `errors` is ignored unless using Python 3.9 or later.
+    """
+
+    def __init__(  # noqa: PLR0913
+        self,
+        filename,
+        when="h",
+        interval=1,
+        backupCount=0,
+        encoding=None,
+        delay=False,
+        utc=False,
+        atTime=None,
+        errors=None,
+        maxBytes=0,
+        use_gzip=False,
+        owner=None,
+        chmod=None,
+        umask=None,
+        newline=None,
+        terminator="\n",
+        unicode_error_policy="ignore",
+        lock_file_directory=None,
+        **kwargs,
+    ):
+        if "mode" in kwargs:
+            del kwargs["mode"]
+        trfh_kwargs = {}
+        if sys.version_info >= (3, 9):
+            trfh_kwargs["errors"] = errors
+        TimedRotatingFileHandler.__init__(
+            self,
+            filename,
+            when=when,
+            interval=interval,
+            backupCount=backupCount,
+            encoding=encoding,
+            delay=delay,
+            utc=utc,
+            atTime=atTime,
+            **trfh_kwargs,
+        )
+        self.clh = ConcurrentRotatingFileHandler(
+            filename,
+            mode="a",
+            backupCount=backupCount,
+            encoding=encoding,
+            delay=delay,
+            maxBytes=maxBytes,
+            use_gzip=use_gzip,
+            owner=owner,
+            chmod=chmod,
+            umask=umask,
+            newline=newline,
+            terminator=terminator,
+            unicode_error_policy=unicode_error_policy,
+            lock_file_directory=lock_file_directory,
+            **kwargs,
+        )
+        self.num_rollovers = 0
+        self.__internal_close()
+        self.initialize_rollover_time()
+
+    def __internal_close(self):
+        # Don't need or want to hold the main logfile handle open unless we're actively writing to it.
+        if self.stream:
+            self.stream.close()
+            # noinspection PyTypeChecker
+            self.stream = None
+
+    def _console_log(self, msg, stack=False):
+        self.clh._console_log(msg, stack=stack)
+
+    def emit(self, record):
+        """
+        Emit a record.
+
+        Override from parent class to handle file locking for the duration of rollover and write.
+        This also does the formatting *before* locks are obtained, in case the format itself does
+        logging calls from within. Rollover also occurs while the lock is held.
+        """
+        try:
+            msg = self.format(record)
+            try:
+                self.clh._do_lock()
+
+                try:
+                    if self.shouldRollover(record):
+                        self.doRollover()
+                except Exception as e:
+                    self._console_log(
+                        "Unable to do rollover: {}\n{}".format(
+                            e, traceback.format_exc()
+                        )
+                    )
+                    # time.sleep(1000)
+
+                self.clh.do_write(msg)
+
+            finally:
+                self.clh._do_unlock()
+        except (KeyboardInterrupt, SystemExit):
+            raise
+        except Exception:
+            self.handleError(record)
+
+    def read_rollover_time(self):
+        # Lock must be held before calling this method
+        lock_file = self.clh.stream_lock
+        if not lock_file or not self.clh.is_locked:
+            # Lock is not being held?
+            self._console_log(
+                "No rollover time (lock) file to read from. Lock is not held?"
+            )
+            return
+        try:
+            lock_file.seek(0)
+            raw_time = lock_file.read()
+        except OSError:
+            self.rolloverAt = 0
+            self._console_log(f"Couldn't read rollover time from file {lock_file!r}")
+            return
+        try:
+            self.rolloverAt = int(raw_time.strip())
+            # self._console_log(
+            #     f"Read rollover time: {self.rolloverAt} - raw: {raw_time!r}"
+            # )
+        except ValueError:
+            self.rolloverAt = 0
+            self._console_log(f"Couldn't read rollover time from file: {raw_time!r}")
+
+    def write_rollover_time(self):
+        """Write the next rollover time (current value of self.rolloverAt) to the lock file."""
+        lock_file = self.clh.stream_lock
+        if not lock_file or not self.clh.is_locked:
+            self._console_log(
+                "No rollover time (lock) file to write to. Lock is not held?"
+            )
+            return
+        lock_file.seek(0)
+        lock_file.write(str(self.rolloverAt))
+        lock_file.truncate()
+        lock_file.flush()
+        os.fsync(lock_file.fileno())
+        self._console_log(f"Wrote rollover time: {self.rolloverAt}")
+
+    def initialize_rollover_time(self):
+        """Run by the __init__ to read an existing rollover time from the lockfile,
+        and if it can't do that, compute and write a new one."""
+        try:
+            self.clh._do_lock()
+            self.read_rollover_time()
+            self._console_log(f"Initializing; reading rollover time: {self.rolloverAt}")
+            if self.rolloverAt != 0:
+                return
+            current_time = int(time.time())
+            new_rollover_at = self.computeRollover(current_time)
+            while new_rollover_at <= current_time:
+                new_rollover_at += self.interval
+            self.rolloverAt = new_rollover_at
+            self.write_rollover_time()
+
+            self._console_log(f"Set initial rollover time: {self.rolloverAt}")
+        finally:
+            self.clh._do_unlock()
+
+    def shouldRollover(self, record):
+        """Determine if the rollover should occur."""
+        # Read the latest rollover time from the file
+        self.read_rollover_time()
+
+        do_rollover = False
+        if super(ConcurrentTimedRotatingFileHandler, self).shouldRollover(record):
+            self._console_log("Rolling over because of time")
+            do_rollover = True
+        elif self.clh.shouldRollover(record):
+            self.clh._console_log("Rolling over because of size")
+            do_rollover = True
+        if do_rollover:
+            return True
+        return False
+
+    def doRollover(self):  # noqa: C901, PLR0912
+        """
+        do a rollover; in this case, a date/time stamp is appended to the filename
+        when the rollover happens.  However, you want the file to be named for the
+        start of the interval, not the current time.  If there is a backup count,
+        then we have to get a list of matching filenames, sort them and remove
+        the one with the oldest suffix.
+
+        This code was adapted from the TimedRotatingFileHandler class from Python 3.11.
+        """
+        self.clh._close()
+        self.__internal_close()
+
+        # get the time that this sequence started at and make it a TimeTuple
+        currentTime = int(time.time())
+        dstNow = time.localtime(currentTime)[-1]
+        t = self.rolloverAt - self.interval
+        if self.utc:
+            timeTuple = time.gmtime(t)
+        else:
+            timeTuple = time.localtime(t)
+            dstThen = timeTuple[-1]
+            if dstNow != dstThen:
+                addend = 3600 if dstNow else -3600
+                timeTuple = time.localtime(t + addend)
+        dfn = self.rotation_filename(
+            self.baseFilename + "." + time.strftime(self.suffix, timeTuple)
+        )
+
+        gzip_ext = ".gz" if self.clh.use_gzip else ""
+
+        counter = 0
+        while os.path.exists(dfn + gzip_ext):
+            counter += 1
+            ending = f".{counter - 1}{gzip_ext}"
+            if dfn.endswith(ending):
+                dfn = dfn[: -len(ending)]
+            dfn = f"{dfn}.{counter}"
+
+        # if os.path.exists(dfn):
+        #     os.remove(dfn)
+
+        self.rotate(self.baseFilename, dfn)
+
+        if self.clh.use_gzip:
+            self.clh.do_gzip(dfn)
+
+        if self.backupCount > 0:
+            for s in self.getFilesToDelete():
+                os.remove(s + gzip_ext)
+        # if not self.delay:
+        #     self.stream = self._open()
+        newRolloverAt = self.computeRollover(currentTime)
+        while newRolloverAt <= currentTime:
+            newRolloverAt = newRolloverAt + self.interval
+        # If DST changes and midnight or weekly rollover, adjust for this.
+        if (self.when == "MIDNIGHT" or self.when.startswith("W")) and not self.utc:
+            dstAtRollover = time.localtime(newRolloverAt)[-1]
+            if dstNow != dstAtRollover:
+                if not dstNow:  # noqa: SIM108
+                    # DST kicks in before next rollover, so we need to deduct an hour
+                    addend = -3600
+                else:
+                    # DST bows out before next rollover, so we need to add an hour
+                    addend = 3600
+                newRolloverAt += addend
+        self.num_rollovers += 1
+        self.rolloverAt = newRolloverAt
+        self.write_rollover_time()
+        self._console_log(f"Rotation completed (on time) {dfn}")
+
+
+# Publish these classes to the "logging.handlers" module, so they can be used
 # from a logging config file via logging.config.fileConfig().
 import logging.handlers  # noqa: E402
 
 logging.handlers.ConcurrentRotatingFileHandler = ConcurrentRotatingFileHandler
+logging.handlers.ConcurrentTimedRotatingFileHandler = ConcurrentTimedRotatingFileHandler
